"""
Node.py

 A Class Representing a Node in a Directed Graph
 
 Author: Ben Kybartas
 Student ID: 260477933
 Date: April 10, 2012
"""
from Edge import *

class Node:
	
	"""The initialization function
	
	Args:
		name (str): The name of our node
		attributes (dict): A dictionary of our attributes, and their corresponding values

	"""
	def __init__(self, name, attributes):
		
		self._name = name
		self._attributes = attributes
		self._incoming_edges = []
		self._outgoing_edges = []
		
	"""Get the name of our node

	Returns:
		(str): The name of the node
	"""
	def get_name(self):
		return self._name
		
	"""Get our incoming edges
	
	Returns:
		(array): The incoming edges
	"""
	def get_incoming_edges(self):
		return self._incoming_edges
	
	"""Get our outgoing edges
	
	Returns:
		(array): The outgoing edges
	"""
	def get_outgoing_edges(self):
		return self._outgoing_edges
	
	"""Switch the Ownership of a node
	
	Params:
		owner(Node): The new owner
		reason(str): The reason for the ownership change
		
	Returns:
		(array): The edges to remove
		(array): The edges to add
	"""
	def new_owner(self, owner, reason):
		
		#We want to remove the previous edges
		#of this node
		to_remove = []
		for edge in self._incoming_edges:
			from_node = edge.get_from_node()
			from_node.remove_outgoing_edge(edge)
			to_remove.append(edge)
		
		#Reset the incoming edges
		self._incoming_edges = []
		
		#Create the new ownership edge
		new_edge = Edge({"Owns" : reason}, self, owner)
		
		#Link the new edge to our other nodes
		self._incoming_edges.append(new_edge)
		owner.add_outgoing_edge(new_edge)
		
		return to_remove, [new_edge]

	"""Kill A Node
	
	Params:
		murderer(Node): The murderer of the node
		
	Returns:
		(array): The edges to remove
		(array): The edges to add
	"""		
	def murder(self, murderer):

		#We want to remove the outgoing relations
		#of this node
		to_remove = []
		for edge in self._outgoing_edges:
			to_node = edge.get_to_node()
			to_node.remove_incoming_edge(edge)
			to_remove.append(edge)
			
		#Reset the outgoing edges
		self._outgoing_edges = []
		
		#We want to add new relations for the murderer
		to_add = []
		
		#The cause for these relations
		cause = "Murder_of_" + self._name
		
		#Check the incoming relations
		for edge in self._incoming_edges:
			attribute = edge.get_name()
			#If the character was friends with, or loved by, another entity
			#then they should develop hatred towards the murderer
			if attribute.keys()[0] == "Loves" or attribute.keys()[0] == "Friends":
				from_node = edge.get_from_node()
				new_edge = Edge({"Hates" : cause}, from_node, murderer)
				from_node.add_outgoing_edge(new_edge)
				murderer.add_incoming_edge(new_edge)
				to_add.append(new_edge)
			
			#If the character was hated by another entity
			#then they should develop frienship towards the murderer
			if attribute.keys()[0] == "Hates":
				from_node = edge.get_from_node()
				new_edge = Edge({"Friends" : cause}, from_node, murderer)
				from_node.add_outgoing_edge(new_edge)
				murderer.add_incoming_edge(new_edge)
				to_add.append(new_edge)				
				
		return to_remove, to_add
		
	"""Set the name of our node

	Args:
		new_name(str): The new name of the node
	"""
	def set_name(self, new_name):
		self._name = new_name
	
	"""Check the value of a given attribute
	
	Args:
		attribute_name (str): The name of the attribute we are searching for
		
	Raises:
		KeyError
	"""
	def check_attribute(self, attribute_name):
		
		return self._attributes[attribute_name]
	
	"""Modify an Attribute to be a new value
	
	Args:
		attribute_name (str): The name of the attribute we are changing
		new_value (obj): The new value of our attribute
	"""
	def modify_attribute(self, attribute_name, new_value):
		self._attributes[attribute_name] = new_value
		
	"""Get the attribute dictionary
	
	Returns:
		(dict): The attribute dictionary
	"""
	def get_attributes(self):
		return self._attributes
				
	"""Add an Attribute to our attributes
	
	Args:
		attribute_name (str): The name of the attribute we are changing
		value (obj): The new value of our attribute
	"""
	def add_attribute(self, attribute_name, value):
		self._attributes[attribute_name] = value
	
	"""Add a single incoming edge
	
	Args:
		incoming_edge (Edge): The new incoming edge
	"""	
	def add_incoming_edge(self, incoming_edge):
		self._incoming_edges.append(incoming_edge)

	"""Add a single outgoing edge
	
	Args:
		outgoing_edge (Edge): The new outgoing edge
	"""			
	def add_outgoing_edge(self, outgoing_edge):
		self._outgoing_edges.append(outgoing_edge)
	
	"""Add an array of incoming edges
	
	Args:
		incoming_edges (array): An array of new incoming edges
	"""	
	def add_incoming_edge_array(self, incoming_edges):
		self._incoming_edges.extend(incoming_edge)

	"""Add an array of outgoing edges
	
	Args:
		outgoing_edges (array): An array of new outgoing edges
	"""	
	def add_outgoing_edge_array(self, outgoing_edges):
		self._outgoing_edges.extend(outgoing_edges)
	
	"""Remove an outgoing edge
	
	Args:
		edge_to_remove (array): The edge to remove
		
	Raises:
		ValueError
	"""	
	def remove_outgoing_edge(self, edge_to_remove):
		self._outgoing_edges.remove(edge_to_remove)

	"""Remove an incoming edge
	
	Args:
		edge_to_remove (array): The edge to remove
		
	Raises:
		ValueError
	"""		
	def remove_incoming_edge(self, edge_to_remove):
		self._incoming_edges.remove(edge_to_remove)
		
	"""Get an incoming edge based off its position in the array
	
	Args:
		edge_id (array): The position of the edge we are retrieving
		
	Returns:
		The edge at position edge_id in the array
		
	Raises:
		IndexError
	"""			
	def get_incoming_edge_from_id(self, edge_id):
		return self._incoming_edges[edge_id]
		
	"""Get an outgoing edge based off its position in the array
	
	Args:
		edge_id (int): The position of the edge we are retrieving
		
	Returns:
		The edge at position edge_id in the array
		
	Raises:
		IndexError
	"""					
	def get_outgoing_edge_from_id(self, edge_id):
		return self._outgoing_edges[edge_id]

	"""Get the node that has an incoming relation to this node
	
	Args:
		edge_name (Edge): The edge which we are searching for
	Returns:
		(Node): The node from the incoming edge
	"""
	def get_from_node_from_incoming_edge_name(self, edge_name):
		for edge in self._incoming_edges:
			if edge.name_equals(edge_name):
				return edge.get_from_node()
		return None

	"""Get the node that has an outgoing relation from this node
	
	Args:
		edge_name (Edge): The edge which we are searching for
	Returns:
		(Node): The node from the outgoing edge
	"""
	def get_to_node_from_outgoing_edge_name(self, edge_name):
		for edge in self._outgoing_edges:
			if edge.name_equals(edge_name):
				return edge.get_to_node()
		return None
		
	"""Check if we are linked to a specific node
	Args:
		node (Node): The node which we are checking
	Returns:
		(Bool): Are we linked to the node?
		(edge): The edge which connects us to the node
	"""
	def connected_to(self, node):
		for edge in self._outgoing_edges:
			if edge.get_to_node() == node:
				return True, edge
		return False, None

	"""Check if we are linked from a specific node
	Args:
		node (Node): The node which we are checking
	Returns:
		(Bool): Are we linked from the node?
		(edge): The edge which connects us from the node
	"""		
	def connected_from(self, node):
		for edge in self._incoming_edges:
			if edge.get_from_node() == node:
				return True, edge
		return False, None
		
	"""Check if a particular node is a sub_node of ourselves
	
	Param:
		sub_node(Node): The sub_node we are checking against
	Returns:
		(Bool): True if it is a sub_node, false otherwise
	"""
	def contains_subnode(self, sub_node):

		#Compare our attributes
		for attr in sub_node.get_attributes():
			
			#If any attributes are not contained, return false
			if attr in self._attributes:
				
				#If any attributes don't match, return false
				if (not self._attributes[attr] == "N/A") and (not self._attributes[attr] == sub_node.get_attributes()[attr]):
					return False
			else:
				return False
		
		#We don't want to recheck the same edge, so have a found_edges array
		found_edges = []
		
		#Compare our incoming edges
		for sub_edge in sub_node.get_incoming_edges():
			has_edge = False
			
			#Check if we have all the same incoming edges
			for edge in self._incoming_edges:
				
				#Check the attribute name of the edge and whether or not we have already seen this edge
				has_edge = has_edge or ((sub_edge.name_equals(edge.get_name())) and (not edge in found_edges)) 
				if has_edge:
					#If we've found the edge, break
					found_edges.append(edge)
					break
			#If we don't find any edges, return false
			if has_edge == False:
				return False
				
		#We don't want to recheck the same edge, so have a found_edges array
		found_edges = []
		
		#Compare our outgoing edges
		for sub_edge in sub_node.get_outgoing_edges():
			has_edge = False
			
			#Check if we have all the same outgoing edges
			for edge in self._outgoing_edges:
				
				#Check the attribute name of the edge and whether or not we have already seen this edge
				has_edge = has_edge or ((sub_edge.name_equals(edge.get_name())) and (not edge in found_edges))
				if has_edge:
					#If we've found the edge, break
					found_edges.append(edge)
					break
			#If we don't find any edges, return false
			if has_edge == False:
				return False

		#If we have matching attributes and edges, return true
		return True
	
