"""
Defines our Scheduler for Narrative Generation

By: Ben Kybartas
Student ID: 260477933

"""

from Node import *
from Edge import *
from Graph import *
from StoryGraph import *
from StoryNode import *
from SocialGraph import *
from random import *

class Scheduler:

	"""Initialization Function
	
	Params:
		social_graph (SocialGraph): Our Game world graph
		story_initialization_rules (RewriteRule): Our narrative skeleton generation
		rewrite_rules (RewriteRule): Our narrative rewrite rules
		social_rewrite_rules (RewriteRule): Our social rewrite rules
	"""
	def __init__(self, social_graph, story_initialization_rules, rewrite_rules, social_rewrite_rules):
		self._story_init_rules = story_initialization_rules
		self._rewrite_rules = rewrite_rules
		self._social_graph = social_graph
	
	"""
	Initialize Our Narrative (Generate a "Skeleton")
	"""
	def initialize_narrative(self):
		
		#Create our starting ending node
		start_node = StoryNode("Start_Quest", {"Node_Type" : "Start"}, "N/A")
		end_node = StoryNode("End_Quest", {"Node_Type" : "End"}, "N/A")

		#Create an empty narrative
		self._narrative = StoryGraph("New_Quest", [start_node, end_node])
		
		#Get all possible narratives that can be generated
		matching_rules = []
		for rule in self._story_init_rules:
			if not rule.get_social_condition() == None:
				
				#Check for Results
				social_results = self._social_graph.contains_subgraph(rule.get_social_condition())
			
			#If we get results, add them to our array
			if len(social_results) > 0:
				matching_rules.append([rule, social_results])
		
		#Return False if no Narratives are possible
		if len(matching_rules) <= 0:
			return False
			
		#Rules for picking the matching narrative
		chosen = randint(0, len(matching_rules) - 1)
		resultant = matching_rules[chosen]
		
		resulting_rule = resultant[0]

		#Rules for picking the matching group
		chosen = randint(0, len(resultant[1]) - 1)
		resulting_group = resultant[1][chosen]

		#Modify our story graph
		story_modification = resulting_rule.get_story_modification()
		
		#Make the nodes
		for node in story_modification.get_nodes():
			index = resulting_rule.get_social_condition().get_nodes().index(node.get_target())
			target = resulting_group[index]
			node.add_attribute("Target", target.get_name())
			node.set_linked_to_node(target)
			self._narrative.add_node(node)
		
		#Make the edges
		for edge in story_modification.get_edges():
			self._narrative.add_edge(edge)
		
		#Connect the starting and ending nodes
		self._narrative.connect(self._narrative.get_start_node(), {}, self._narrative.get_nodes()[2])
		self._narrative.connect(self._narrative.get_nodes()[len(self._narrative.get_nodes()) -1 ], {}, self._narrative.get_end_node())
		
		#Plot our Starting Narrative
		self._narrative.plot('Starting_Narrative')
		return True
		
		
	def write_narrative(self):
		self._final_narrative = StoryGraph("My_Narrative", [])
		num_of_refinements = randint(1, 5)
		
		self.get_story_trace(self._narrative.get_nodes()[0], self._narrative.get_nodes(), num_of_refinements)
		
		self._final_narrative.plot('Final_Narrative')
	
	#Simulate Our Story
	def get_story_trace(self, starting_node, node_list, refinements_remaining):
		
		#Our previous and current nodes are initalized to be the same
		prev_node = starting_node
		cur_node = starting_node
		
		#Begin making our narrative
		self._final_narrative.add_node(cur_node)
		
		#While we are not at an ending node
		while len(cur_node.get_outgoing_edges()) > 0:
			
			#Get the next event
			cur_node = cur_node.get_outgoing_edges()[0].get_to_node()
			
			#Add the new event to our narrative
			self._final_narrative.add_node(cur_node)

			#Get the social modification
			modify = cur_node.get_modification()
			
			#If there is a modification
			if not modify == None:
				
				#Get the nodes to modify
				target_modify = modify.get_nodes()[0]
				
				#Modify their attributes
				for attribute in target_modify.get_attributes():
					cur_node.get_linked_to_node().modify_attribute(attribute, target_modify.get_attributes()[attribute])
					
					#If we are murdering someone, run the murder rule (see Node)
					if attribute == "alive" and target_modify.get_attributes()[attribute] == False:
						to_remove, to_add = cur_node.get_linked_to_node().murder(self._social_graph.get_player())
						self._social_graph.delete_edges(to_remove)
						self._social_graph.add_edges(to_add)
						self._social_graph.update_adj()
						
					#If we are stealing something, run the stealing rule (see Node)
					if attribute == "status" and target_modify.get_attributes()[attribute] == "stolen":
						to_remove, to_add = cur_node.get_linked_to_node().new_owner(self._social_graph.get_player(), "stolen")
						self._social_graph.delete_edges(to_remove)
						self._social_graph.add_edges(to_add)
						self._social_graph.update_adj()
			
			#If there is a previous node
			if not prev_node == None:
				
				applied_rule = False
				
				#Connect our previous node to our current node
				self._final_narrative.connect(prev_node, {}, cur_node)
				
				#Go through our rewrite rules and see if any match
				for rule in self._rewrite_rules:
					if not applied_rule:
						
						#Compare our story conditions
						story_condition = rule.get_story_condition().get_nodes()[0]
						
						#If the current node is a sub node and we 
						#haven't refined our story too many times
						#add the story twist 
						if cur_node.contains_subnode(story_condition) and refinements_remaining > 0:
							
							#Now we check social condition
							social_condition = rule.get_social_condition()
							
							#Get the nodes
							nodes = social_condition.get_nodes()
							target_node = nodes[0]

							#Check if the social condition is valid
							if cur_node.get_linked_to_node().contains_subnode(target_node):
								
								#Get the new target
								new_target = cur_node.get_linked_to_node().get_from_node_from_incoming_edge_name(target_node.get_incoming_edges()[0].get_name())
								
								#Get the story modification
								results = rule.get_story_modification()

								#Apply the story modification to our narrative
								for node in results.get_nodes():
									if not node == story_condition:
										new_node = StoryNode(node.get_name(), node.get_attributes(), new_target)
										new_node.add_attribute("Target", new_target.get_name())
										new_node.set_linked_to_node(new_target)
										
										#Check if we need to make a game world modification
										modify = node.get_modification()
										
										#Apply the modification if necessary
										if not modify == None:
											target_modify = modify.get_nodes()[0]
											for attribute in target_modify.get_attributes():
												new_node.get_linked_to_node().modify_attribute(attribute, target_modify.get_attributes()[attribute])
												
												#If we have murdered someone, apply the murder rule (see Node)
												if attribute == "alive" and target_modify.get_attributes()[attribute] == False:
													to_remove, to_add = new_node.get_linked_to_node().murder(self._social_graph.get_player())
													self._social_graph.delete_edges(to_remove)
													self._social_graph.add_edges(to_add)
													self._social_graph.update_adj()
													
												#If we have stolen something, apply the stole rule (see Node)
												if attribute == "status" and target_modify.get_attributes()[attribute] == "stolen":
													to_remove, to_add = new_node.get_linked_to_node().new_owner(self._social_graph.get_player(), "stolen")
													self._social_graph.delete_edges(to_remove)
													self._social_graph.add_edges(to_add)
													self._social_graph.update_adj()
													
										#Add the new node to our graph and connect it
										self._final_narrative.add_node(new_node)
										self._final_narrative.connect(cur_node, {}, new_node)
										
										#Set it to be our previous node
										prev_node = new_node
										
										#Reduce the refinements remaining
										refinements_remaining -= 1
										
										#We have applied a rule
										applied_rule = True
										
				#If we did not apply a rule, our new previous
				#node is our current node
				if not applied_rule:
					prev_node = cur_node
			else:
				prev_node = cur_node

